/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */


/**
 * @file reportGenerator.cpp
 * @brief Report generator.
 * @version 1.0.0
 * @date 2024-12-20
 */

 /* === Includes === */
#include "report_generator.h"
#include <algorithm>
#include <cctype>
#include <cstddef>
#include <filesystem>
#include <moduleBasics/module.h>
#include <runtimeInfo/runtimeInfo.h>
#include <stdexcept>
#include <string>
#include <vector>
#include <aircraftGeometry2/airfoil_surface.h>
#include <aircraftGeometry2/fuselage.h>
#include <aircraftGeometry2/hull_surface.h>
#include <aircraftGeometry2/geometry/surface.h>
#include <aircraftGeometry2/geometry/builder.h>
#include <aircraftGeometry2/geometry/entity3d.h>
#include <aircraftGeometry2/geometry/factory.h>
#include <aircraftGeometry2/geometry/section.h>
#include <aircraftGeometry2/io/convert.h>
#include <aircraftGeometry2/io/dat.h>
#include <aircraftGeometry2/processing/transform.h>
#include <aircraftGeometry2/processing/measure.h>

/* === Methods === */

void ReportGenerator::initialize()
{
    std::cout << "Initializing the report generation module." << std::endl;

    /* Get the XML resources */
    this->configuration_xml = aixml::openDocument(this->get_RuntimeIO()->moduleConfAccess);
    this->aircraft_xml = aixml::openDocument(this->get_RuntimeIO()->acxmlAccess);

    /* Define the filename and the path of the template folder */
    std::string filename = "UNICADO-Report-Template.html";
    std::filesystem::path filePath = "./inc/html_templates/" + filename;

    /* Get the report path */
    const std::string project_path(this->configuration_xml->at("module_configuration_file/aircraft_exchange_file_directory/value"));
    this->reporting_path = fs::path(rtIO_->getHtmlDir());

    /* Check if the files exist */
    if (std::filesystem::exists(filePath)) {
        std::cout << "UNICADO-Report-Template.html exists, proceeding" << std::endl;
        this->htmlcontent.open(filePath);
        this->html_temp.open(this->reporting_path / "UNICADO-Report.html", std::ios::out);
    }
    else {
        std::cout << "UNICADO-Report-template.html does not exist" << filePath << std::endl;
    }
    
    /* Copy the style.css file to the reporting folder */
    std::filesystem::copy_file(std::filesystem::path("./inc/html_templates/style-template.css"), this->reporting_path / "style.css", std::filesystem::copy_options::overwrite_existing);

    /* Copy the UNICADO logo to the reporting folder*/
    std::filesystem::copy_file(std::filesystem::path("./inc/logos/UNICADO_logo_refined.svg"), this->reporting_path / "UNICADO_logo_refined.svg", std::filesystem::copy_options::overwrite_existing);
    std::filesystem::copy_file(std::filesystem::path("./inc/logos/UNICADO_logo.svg"), this->reporting_path / "UNICADO_logo.svg", std::filesystem::copy_options::overwrite_existing);

}

void ReportGenerator::run()
{
    /* Get the TLARs */
    this->TLARs = this->aircraft_xml->getVector("aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements");

    /* Write the content to the html report file */
    ReportGenerator::write_dropdown_content();
}

void ReportGenerator::update()
{
    this->htmlcontent.close();
    this->html_temp.close();
}

void ReportGenerator::save()
{

#ifdef _WIN32
    /* Read the batch creation switch from the config file */
    const std::string enable_batch(this->configuration_xml->at("module_configuration_file/control_settings/program_specific_settings/create_batch_file/value"));
    if (enable_batch == "true") {
        /* Create a batch file to start the server to host the report */
        const std::string project_path(this->configuration_xml->at("module_configuration_file/aircraft_exchange_file_directory/value"));
        std::filesystem::path bat_path = std::filesystem::path(project_path) / "reporting/";
        std::fstream bat_file;
        bat_file.open(bat_path / "com.bat", std::ios::out);
        bat_file << "cd " << bat_path.string() << std::endl;
        bat_file << "start cmd.exe @cmd /k \"python -m http.server\"" << std::endl;
        bat_file << "start \"\" \"http://localhost:8000/report_html/UNICADO-Report.html\"";
        bat_file.close();
        bat_path /= "com.bat";
        std::cout << "initializing the report -> " << bat_path.string().c_str() << std::endl;
        std::string full_path = bat_path.string();
        std::replace(full_path.begin(), full_path.end(), '/', '\\');
        std::string command = "cmd /k " + full_path;
        std::system(command.c_str());
    }
#endif
}

void ReportGenerator::report()
{
    /* Update the html file */
    std::cout << "Report generation completed" << std::endl;

}

void ReportGenerator::get_all_TLAR(const auto TLAR)
{
    /* Store the standard parameters that will be displayed at startup */
    std::vector<std::string> def_TLAR = { "range", "maximum_structrual_payload_mass", "initial_cruise_mach_number", "maximum_operating_mach_number", "initial_cruise_altitude", "time_to_climb", "maximum_one_engine_operating_altitude", "maximum_operating_altitude", "takeoff_distance", "landing_field_lengt", "maximum_approach_speed" };
    
    /* Find the value, upper, lower boundaries of the TLAR recursively */
    if ((TLAR->getChildren()).size() > 0) {
        for (const auto& childTLAR : TLAR->getChildren())
        {
            if (childTLAR->name != "value" && childTLAR->name != "unit" && childTLAR->name != "upper_boundary" && childTLAR->name != "lower_boundary") {
                ReportGenerator::get_all_TLAR(childTLAR);
            }
        }
    }

    /* Populate the label section of the html file, if TLAR is in the def_TLAR vector, mark the checkbox */
    try 
    {
        auto TLAR_name =  TLAR->name;
        auto TLAR_value = TLAR->at("/value");
        auto TLAR_unit = TLAR->at("/unit");
        auto TLAR_lb = TLAR->at("/lower_boundary");
        auto TLAR_ub = TLAR->at("/upper_boundary");
        std::string label;
        if (std::find(def_TLAR.begin(), def_TLAR.end(), TLAR_name) != def_TLAR.end()) {
            label = "<label><input type = ""checkbox"" value = \"" + ReportGenerator::format_element(TLAR_name) + "\" data-value=""" + TLAR_value + """ data-unit=""" + TLAR_unit + """ data-lower=""" + TLAR_lb + """ data-upper=""" + TLAR_ub + """ checked> " + TLAR_name + "</label>";
        }
        else {
            label = "<label><input type = ""checkbox"" value = \"" + ReportGenerator::format_element(TLAR_name) + "\" data-value=""" + TLAR_value + """ data-unit=""" + TLAR_unit + """ data-lower=""" + TLAR_lb + """ data-upper=""" + TLAR_ub + """> " + TLAR_name + "</label>";
        }
        this->html_temp << label << std::endl;
    }
    catch (...)
    {
        auto TLAR_name = TLAR->name;
        auto TLAR_value = TLAR->at("/value");
        std::string label = "<label><input type = ""checkbox"" value = \"" + ReportGenerator::format_element(TLAR_name) + "\" data-value=""" + TLAR_value + """ data-unit=""" + "n/a" + """ data-lower=""" + "n/a" + """ data-upper=""" + "n/a" + """> " + TLAR_name + "</label>";
        this->html_temp << label << std::endl;
    }

}

void ReportGenerator::write_dropdown_content()
{
    if (this->htmlcontent) {
        for (const auto& TLAR : this->TLARs)
        {
            /* Read the line of the template report file  */
            std::string line;
            while (getline(this->htmlcontent, line))
            {
                /* Write the line to the report  */
                this->html_temp << line << std::endl;

                /* If the line is the starting point of the dropdown content, write the TLARs  */
                if (line.find("<div id=\"dropdownContent\" class=\"dropdown-content\">") != std::string::npos)
                {
                    ReportGenerator::get_all_TLAR(TLAR);
                    this->html_temp << "</div>" << std::endl;
                    while (line.find("</div>") == std::string::npos) { getline(this->htmlcontent, line); }
                }
                /* If the line is the starting point of the linked reports, add links to all reports in the folder */
                if (line.find("<ul id = \"list of linked reports\">") != std::string::npos)
                {
                    ReportGenerator::add_links(this->reporting_path);
                    this->html_temp << "</ul>" << std::endl;
                    while (line.find("</ul>") == std::string::npos) { getline(this->htmlcontent, line); }
                }
                /* If the line is the starting point of the aircraft charactersitics data structure, populate it */
                if (line.find("const clusters = {") != std::string::npos)
                {
                    ReportGenerator::add_clusters();
                    std::cout << "finished" << std::endl;
                }
            }
        }
    }
    /* Create the aircraft geometry to be displayed in the report */
    ReportGenerator::create_aircraft_geometry();
}

auto ReportGenerator::format_element(std::string element) -> std::string 
{
    /* Replace underscore with space */
    /* @todo make each letter after a space upper-case */
    std::replace(element.begin(), element.end(), '_', ' ');
    element[0] = std::toupper(element[0]);
    return element;
}

void ReportGenerator::add_links(std::filesystem::path reporting_path) 
{   
    /* Search the reporting directory for report.html files */
    for (const auto& entry : std::filesystem::directory_iterator(reporting_path)) {
        if (entry.path().string().find("_report.html") != std::string::npos) {
            std::string path_entry = entry.path().string();
            std::replace(path_entry.begin(), path_entry.end(), '\\', '/');
            unsigned first = path_entry.find_last_of("/");
            unsigned last = path_entry.find("_report.html");
            /* Add the link to the report */
            std::string report_title = path_entry.substr(first + 1, last - first + 6);
            this->html_temp << "<li><a href = \"" + path_entry.substr(first + 1) + "\" target = \"_blank\">" + ReportGenerator::format_element(report_title) + " </a> </li>"
                            << std::endl;
        }
    }
}

void ReportGenerator::add_clusters()
{
    std::string line;
    int entry = 0;
    while ((getline(this->htmlcontent, line)) && (entry < 6))
    {
        this->html_temp << line << std::endl;
        /* Search the aircraft xml file for paths to clustered aircraft design parameters */
        /* Write to the cluster if found */
        try
        {
            if (line.find("assessment: [") != std::string::npos)
            {
                entry += 1;
                this->html_temp << "{ item: \"takeoff distance normal safety\", " +
                    ReportGenerator::generic_writer("aircraft_exchange_file/assessment/performance/takeoff/takeoff_distance_normal_safety") + "," << std::endl;
                this->html_temp << "{ item: \"needed runway length\"," + 
                    ReportGenerator::generic_writer("assessment/performance/landing/needed_runway_length") << std::endl;
            }
            if (line.find("sizingPoint: [") != std::string::npos)
            {
                entry += 1;
                this->html_temp << "{ item: \"thrust to weight ratio\", " +
                    ReportGenerator::generic_writer("sizing_point/thrust_to_weight") + "," << std::endl;
                this->html_temp << "{ item: \"wing loading\"," +
                    ReportGenerator::generic_writer("sizing_point/wing_loading") << std::endl;
            }
            if (line.find("liftCoefficients: [") != std::string::npos)
            {
                entry += 1;
                this->html_temp << "{ item: \"CL optimum cruise\", " +
                    ReportGenerator::generic_writer("analysis/aerodynamics/lift_coefficients/C_LoptimumCruise") + "," << std::endl;
                this->html_temp << "{ item: \"CL max landing\"," +
                    ReportGenerator::generic_writer("analysis/aerodynamics/lift_coefficients/C_LmaxLanding") + "," << std::endl;
                this->html_temp << "{ item: \"CL max take-off\", " +
                    ReportGenerator::generic_writer("analysis/aerodynamics/lift_coefficients/C_LmaxT-O") + "," << std::endl;
                this->html_temp << "{ item: \"CL ground roll\"," +
                    ReportGenerator::generic_writer("analysis/aerodynamics/lift_coefficients/C_LgroundRoll") << std::endl;
            }
            if (line.find("aerodynamics: [") != std::string::npos)
            {
                entry += 1;
                this->html_temp << "{ item: \"span\", " +
                    ReportGenerator::generic_writer("analysis/aerodynamics/reference_values/b") + "," << std::endl;
                this->html_temp << "{ item: \"mean aerodynamic chord\"," +
                    ReportGenerator::generic_writer("analysis/aerodynamics/reference_values/MAC") + "," << std::endl;
                this->html_temp << "{ item: \"wing area\", " +
                    ReportGenerator::generic_writer("analysis/aerodynamics/reference_values/S_ref") + "," << std::endl;
            }
            if (line.find("masses: [") != std::string::npos)
            {
                entry += 1;
                this->html_temp << "{ item: \"maximum takeoff mass\", " +
                    ReportGenerator::generic_writer("analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/mass") + "," << std::endl;
                this->html_temp << "{ item: \"operating mass empty\"," +
                    ReportGenerator::generic_writer("analysis/masses_cg_inertia/operating_mass_empty/mass_properties/mass") + "," << std::endl;
                this->html_temp << "{ item: \"design mass\", " +
                    ReportGenerator::generic_writer("analysis/masses_cg_inertia/design_mass/mass_properties/mass") + "," << std::endl;
                this->html_temp << "{ item: \"manufacturer mass empty\"," +
                    ReportGenerator::generic_writer("analysis/masses_cg_inertia/manufacturer_mass_empty/mass_properties/mass") + "," << std::endl;
                this->html_temp << "{ item: \"zero fuel mass\", " +
                    ReportGenerator::generic_writer("analysis/masses_cg_inertia/maximum_zero_fuel_mass/mass_properties/mass") + "," << std::endl;
                this->html_temp << "{ item: \"maximum payload mass\", " +
                    ReportGenerator::generic_writer("analysis/masses_cg_inertia/maximum_payload_mass/mass_properties/mass") + "," << std::endl;
                this->html_temp << "{ item: \"maximum fuel mass\"," +
                    ReportGenerator::generic_writer("analysis/masses_cg_inertia/maximum_fuel_mass/mass_properties/mass") << std::endl;
            }
            if (line.find("cost: [") != std::string::npos)
            {
                entry += 1;
                this->html_temp << "{ item: \"direct operating cost\"," +
                    ReportGenerator::generic_writer("assessment/cost_estimation/operating_cost/direct_operating_cost") << std::endl;
            }
        }
        /* If cluster contains missing elements, inform the user */
        catch (...)
        {
            switch (entry) 
            {
            case 0:
                std::cout << "Key aircraft parameter cluster" << "assessment" << " incomplete" << std::endl;
            case 1:
                std::cout << "Key aircraft parameter cluster" << "sizingPoint" << " incomplete" << std::endl;
            case 2:
                std::cout << "Key aircraft parameter cluster" << "liftCoefficients" << " incomplete" << std::endl;
            case 4:
                std::cout << "Key aircraft parameter cluster" << "aerodynamics" << " incomplete" << std::endl;
            case 5:
                std::cout << "Key aircraft parameter cluster" << "masses" << " incomplete" << std::endl;
            }
        }
        while (line.find("]") == std::string::npos) { getline(this->htmlcontent, line); }
        this->html_temp << line << std::endl;
    }
    this->html_temp << line << std::endl;
}

auto ReportGenerator::generic_writer(std::string element) -> std::string
{
    std::string label;
    /* Write the value and unit of a given node */
    try
    {
        const std::string value(this->aircraft_xml->at(element + "/value"));
        const std::string unit(this->aircraft_xml->at(element + "/unit"));
        label = "value : \"" + value + "\", unit : \"" + unit + "\" }";
    }
    /* Write the value of a given node */
    catch (...)
    {
        const std::string value(this->aircraft_xml->at(element + "/value"));
        label = "value : \"" + value + "\", unit : \"\" }";
    }
    return label;
}

void ReportGenerator::create_aircraft_geometry()
{
    /* Lambda to extract the component position reference */
    auto get_position = [](const node& component) -> geom2::Point_3
        {
            return geom2::Point_3{ double{component.at("position/x/value")},
                double{component.at("position/y/value")},
                double{component.at("position/z/value")} };
        };

    /* Create the aircraft geometry */
    try
    {
        /* Find all aircraft parts in the xml file */
        const auto wings = this->aircraft_xml->getVector("aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface");
        const auto wing_position = get_position(this->aircraft_xml->at("aircraft_exchange_file/component_design/wing"));
        const auto fuselages = this->aircraft_xml->getVector("aircraft_exchange_file/component_design/fuselage/specific/geometry/fuselage");
        const auto fuselage_position = get_position(this->aircraft_xml->at("aircraft_exchange_file/component_design/fuselage"));
        const auto empennage = this->aircraft_xml->getVector("aircraft_exchange_file/component_design/empennage/specific/geometry/aerodynamic_surface");
        const auto empennage_position = get_position(this->aircraft_xml->at("aircraft_exchange_file/component_design/empennage"));
        const auto propulsion = this->aircraft_xml->getVector("aircraft_exchange_file/component_design/propulsion/specific/propulsion");
        auto propulsion_position = get_position(this->aircraft_xml->at("aircraft_exchange_file/component_design/propulsion/"));

        /* Create the surface builders */
        geom2::WingFactory wing_factory{ this->aircraft_xml, this->get_RuntimeIO()->getAirfoilDataDir() };
        geom2::FuselageFactory fuselage_factory{ this->aircraft_xml, this->get_RuntimeIO()->getGeometryDir() };
        geom2::AirfoilSurfaceFactory airfoil_factory{ this->aircraft_xml, this->get_RuntimeIO()->getGeometryDir() };
        geom2::HullFactory hull_factory{ this->aircraft_xml, this->get_RuntimeIO()->getGeometryDir() };

        /* Create the aircraft geometry */
        this->aircraft_geometry = std::make_shared<geometry::Aircraft>();
        std::transform(wings.begin(), wings.end(), std::back_inserter(this->aircraft_geometry->wings),
            [&wing_factory](const auto& wing)
            { return wing_factory.create(
                "wing/specific/geometry/aerodynamic_surface@" + wing->getStringAttrib("ID")); });
        std::transform(fuselages.begin(), fuselages.end(), std::back_inserter(this->aircraft_geometry->fuselages),
            [&fuselage_factory](const auto& fuselage)
            { return fuselage_factory.create(
                "fuselage/specific/geometry/fuselage@" + fuselage->getStringAttrib("ID")); });
        std::transform(empennage.begin(), empennage.end(), std::back_inserter(this->aircraft_geometry->empennage),
            [&wing_factory](const auto& empennage)
            { return wing_factory.create(
                "empennage/specific/geometry/aerodynamic_surface@" + empennage->getStringAttrib("ID")); });
        std::transform(propulsion.begin(), propulsion.end(), std::back_inserter(this->aircraft_geometry->nacelle),
            [&hull_factory](const auto& propulsion)
            { return hull_factory.create(
                "propulsion/specific/propulsion@" + propulsion->getStringAttrib("ID") + "/nacelle@0"); });
        std::transform(propulsion.begin(), propulsion.end(), std::back_inserter(this->aircraft_geometry->pylon),
            [&airfoil_factory](const auto& propulsion)
            { return airfoil_factory.create(
                "propulsion/specific/propulsion@" + propulsion->getStringAttrib("ID") + "/pylon@0"); });

        /* Convert the origins of the wings to the global coordinate system and add them to the mesh */
        auto offset = wing_position - CGAL::ORIGIN;
        geom2::Kernel::Aff_transformation_2 mirror_2d = geom2::Kernel::Aff_transformation_2(1, 0, 0, -1);
        std::ranges::for_each(
            this->aircraft_geometry->wings,
            [&offset, this, mirror_2d](auto& wing)
            { wing.origin += offset;
              this->aircraft_mesh += geom2::transform::to_mesh(wing);
              /* Create symmetical wing if the given wing has the flag if_symmetric = true */
              if (wing.is_symmetric) {
                  geom2::Kernel::Aff_transformation_3 mirror = geom2::Kernel::Aff_transformation_3(1, 0, 0, 0, -1, 0, 0, 0, 1);
                  for (auto& wing_section : wing.sections) {
                      wing_section.normal = mirror(wing_section.normal);
                      wing_section.origin = mirror(wing_section.origin);
                      std::vector<geom2::Point_2> mirrored_contour = {};
                      for (auto& point : wing_section.get_contour()) {
                          mirrored_contour.insert(mirrored_contour.begin(),mirror_2d(point));
                      }
                      auto airfoil = geom2::Polygon_2(mirrored_contour.begin(), mirrored_contour.end());
                      wing_section.set_contour(airfoil);
                  }
                  wing.normal = mirror(wing.normal);
                  this->aircraft_mesh += geom2::transform::to_mesh(wing);
              }
            });

        /* Convert the origins of the fuselages to the global coordinate system and add them to the mesh */
        offset = fuselage_position - CGAL::ORIGIN;
        std::ranges::for_each(
            this->aircraft_geometry->fuselages,
            [&offset, this](auto& fuselage)
            { fuselage.origin += offset;
              this->aircraft_mesh += geom2::transform::to_mesh(fuselage); });

        /* Convert the origins of the empennage to the global coordinate system and add them to the mesh */
        offset = empennage_position - CGAL::ORIGIN;
        std::ranges::for_each(
            this->aircraft_geometry->empennage,
            [&offset, this](auto& empennage)
            { empennage.origin += offset;
              this->aircraft_mesh += geom2::transform::to_mesh(empennage);
              /* Create symmetical empennage if the given empennage has the flag if_symmetric = true */
              if (empennage.is_symmetric) {
                  geom2::Kernel::Aff_transformation_3 mirror = geom2::Kernel::Aff_transformation_3(1, 0, 0, 0, -1, 0, 0, 0, 1);
                  for (auto& empennage_section : empennage.sections) {
                      empennage_section.normal = mirror(empennage_section.normal);
                      empennage_section.origin = mirror(empennage_section.origin);
                  }
                  empennage.normal = mirror(empennage.normal);
                  this->aircraft_mesh += geom2::transform::to_mesh(empennage);
              }
            });

        /* Convert the origins of the engine assemblies to the global coordinate system and add them to the mesh */
        int index = 0;
        for (auto& engine : propulsion) {
            propulsion_position = get_position(this->aircraft_xml->at("propulsion/specific/propulsion@" + engine->getStringAttrib("ID") + "/engine"));
            offset = propulsion_position - CGAL::ORIGIN;
            this->aircraft_geometry->nacelle[index].origin += offset;
            this->aircraft_geometry->pylon[index].origin += offset;
            this->aircraft_mesh += geom2::transform::to_mesh(this->aircraft_geometry->nacelle[index]);
            this->aircraft_mesh += geom2::transform::to_mesh(this->aircraft_geometry->pylon[index]);
            index += 1;
        }

        /* Write the mesh to an stl file */
        std::fstream mesh_file;
        mesh_file.open(this->reporting_path / "AircraftMesh.stl", std::ios::out);
        CGAL::IO::write_STL(mesh_file, this->aircraft_mesh);
        mesh_file.close();
    }
    /* Set the aircraft geometry null if there is a missing element */
    catch (const std::string& exception)
    {
        this->aircraft_geometry = nullptr;
    }
}